
#define IDI_JPEGQS   10

#define IDD_DIALOG   100
#define IDC_LOAD     101
#define IDC_STATUS   102
#define IDC_SAVE     103
#define IDC_OPTIONS  104
#define IDC_FILENAME 105
#define IDC_CONSOLE  106
