#define NO_GETENV
#ifndef XMD_H
#define XMD_H
#include <stdint.h>
typedef int16_t INT16;
typedef int32_t INT32;
#endif

#define HAVE_PROTOTYPES
#define HAVE_UNSIGNED_CHAR
#define HAVE_UNSIGNED_SHORT
#define HAVE_STDDEF_H
#define HAVE_STDLIB_H

#ifdef JPEG_INTERNALS
#define INLINE inline
#endif

