
#include "winuser.h"
#include "dialog.h"

IDI_JPEGQS ICON "jpegqs.ico"

#define DlgW 256
#define DlgH 196
#define BR 6
#define B2 2
#define BW 48
#define BH 14
#define LW 30
#define LH 14

IDD_DIALOG DIALOGEX 0, 0, DlgW, DlgH
STYLE DS_SETFONT | DS_SHELLFONT | DS_MODALFRAME | DS_CENTER | DS_3DLOOK | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "JPEG Quant Smooth"
FONT 8, "MS Shell Dlg"
{
  PUSHBUTTON "Load", IDC_LOAD, BR, BR, BW, BH, WS_DISABLED | WS_TABSTOP
  LTEXT "", IDC_STATUS, BR*2+BW, BR+B2, DlgW-BW*2-BR*4, BH, ES_CENTER
  PUSHBUTTON "Save", IDC_SAVE, DlgW-BW-BR, BR, BW, BH, WS_DISABLED | WS_TABSTOP

  LTEXT "Options:", -1, BR, BR*2+BH+B2, LW, LH
  EDITTEXT IDC_OPTIONS, BR*2+LW, BR*2+BH, DlgW-LW-BR*3, LH, WS_BORDER | WS_TABSTOP

  LTEXT "Filename:", -1, BR, BR*3+BH+LH+B2, LW, LH
  EDITTEXT IDC_FILENAME, BR*2+LW, BR*3+BH+LH, DlgW-LW-BR*3, LH, WS_BORDER | WS_TABSTOP

  EDITTEXT IDC_CONSOLE, BR, BR*4+BH+LH*2, DlgW-BR*2, DlgH-BR*5-BH-LH*2, WS_BORDER | WS_TABSTOP
		| ES_READONLY | ES_MULTILINE | WS_VSCROLL | ES_AUTOVSCROLL | ES_AUTOHSCROLL
}
